# Default settings
default_settings = {
    "mod_expt_file": "data/mod_expt.npy",
    "mod_sld_file": "data/mod_sld_fp49.npy",
    "mod_params_file": "data/mod_params_fp49.npy",

    # hyperparameter settings
    "latent_dim": 2,
    "batch_size": 16,
    "ae_epochs": 20,
    "mlp_epochs": 20,

}

INIT_YAML_CONTENT = f"""\
# 🛠 Configuration file for NR-SLD-Chi Predictor
# Modify these settings according to your project structure.

root: .

### ⚙️ SLD-Chi settings ###
sld_predict_chi:
    file:
        model_experimental_sld_profile: {default_settings["mod_expt_file"]} # 📉 Experimental SLD profile data file (for Chi Prediction)
        model_sld_file: {default_settings["mod_sld_file"]} # 📉 SLD Profile file (input for training)
        model_chi_params_file: {default_settings["mod_params_file"]} # 📉 Chi Parameters file (input label for training)
    
    models:
        latent_dim: {default_settings["latent_dim"]}  # Dimension for latent space
        batch_size: {default_settings["batch_size"]}  # Batch size for training
        ae_epochs: {default_settings["ae_epochs"]}  # Autoencoder training epochs
        mlp_epochs: {default_settings["mlp_epochs"]}  # MLP training epochs

### ⚙️ NR-SLD profile settings ###
nr_predict_sld:
    file:
        nr_train: data/curves/nr_train.npy   #Generated nr curves for training
        sld_train: data/curves/sld_train.npy    #Generated sld curves for training
        
    models:
        model: data/trained_nr_sld_model.pth # Path to save and load the CNN model
        num_film_layers: 5  # Number of film layers to generate the simulated physical model
        num_curves: 50000   # Number of generated curves for training 
        epochs: 10 # CNN training epochs
        batch_size: 32 # Batch size for training
        layers: 12 # model layers
        dropout: 0.5 # Dropout rate for training
        normalization_stats: data/normalization_stat.npy   #Normalization statistics
"""


from .config import (
    NRSLDCurvesGeneratorParams,
    ChiPredTrainingParams,
    NRSLDModelTrainerParams
)
from .chi_pred_model_trainer import ChiPredModelTrainer
from .nr_sld_model_trainer import NRSLDModelTrainer
from .autoencoder import (
    Autoencoder,
    VariationalAutoencoder,
    train as train_ae,
    train_vae
)
from .mlp import deep_MLP as MLP

__all__ = [
    "NRSLDCurvesGeneratorParams",
    "NRSLDModelTrainerParams",
    "ChiPredTrainingParams",
    "ChiPredModelTrainer",
    "NRSLDModelTrainer",
    # "FilmLayer",
    # "FilmLayerBound",
    "Autoencoder",
    "VariationalAutoencoder",
    "train_ae",
    "train_vae",
    "MLP"
]


from abc import abstractmethod
import torch
from ..config.runtime import DEVICE

class ModelTrainer:
    def __init__(self,batch_size, epochs):
        self.batch_size = batch_size
        self.epochs = epochs

    @abstractmethod
    def train_pipeline(self):
        pass

    @staticmethod
    def train_model(model, train_loader, optimizer, loss_fn):
        model.train().to(DEVICE)
        total_loss = 0

        for data, label in train_loader:
            data, label = data.to(DEVICE), label.to(DEVICE)
            optimizer.zero_grad()
            output = model(data)
            loss = loss_fn(output, label)
            loss.backward()
            optimizer.step()
            total_loss += loss.item()

        return total_loss / len(train_loader)

    @staticmethod
    def validate_model(model:torch.nn.Module, valid_loader, loss_fn):
        model.eval().to(DEVICE)
        total_loss = 0

        with torch.no_grad():
            for data, label in valid_loader:
                data, label = data.to(DEVICE), label.to(DEVICE)
                output = model(data)
                loss = loss_fn(output, label)
                total_loss += loss.item()

        return total_loss / len(valid_loader)



from . import train_autoencoder_mlp_chi_pred as chi_prediction_model_trainer

# Selective function imports for direct access
from .run_chi_prediction import run_chi_prediction
from .run_sld_prediction import run_sld_prediction
from .helper import compute_nr_from_sld
__all__ = [
    "compute_nr_from_sld",
    "chi_prediction_model_trainer",  # Full module alias
    "reflectivity_pipeline",  # Full module alias
    "run_chi_prediction",  # Direct function
    "run_sld_prediction",  # Direct function
]


import numpy as np
from refnx.reflect import SLD, ReflectModel
from refl1d.sample.reflectivity import reflectivity_amplitude as reflamp, convolve

def compute_nr_from_sld(
    sld_arr,
    Q=None,
    qmin=0.008,
    qmax=0.20,
    n_q=400,
    # Use ONE of these:
    dq_over_q=None,             # fractional resolution (e.g. 0.0125)
    q_resolution=None,          # matches your generator: dq = (q_resolution/2.355)*Q
    # Roughness:
    sigma=0.0,                  # default 0 for smooth continuous SLD(z)
    # Media handling:
    rho_fronting=None,          # if None, infer from profile front end
    rho_backing=None,           # if None, infer from profile back end
    # Numerics:
    order="substrate_to_air",   # or "air_to_substrate"
    dz=None,                    # e.g. 1.0 Å resampling for stable microslicing
    pad_front=200.0,            # Å plateau padding for semi-infinite fronting
    pad_back=200.0,             # Å plateau padding for semi-infinite backing
    plateau_pts=30,             # number of points to average for inferred front/back SLD
    apply_convolution=True,
):
    """
    Compute neutron reflectivity from a continuous SLD profile using refl1d's slab reflectivity.

    Parameters
    ----------
    sld_arr: (z, rho) or (z, rho, irho)
        z in Å, rho in 1e-6 Å^-2
    order:
        "substrate_to_air" means z increases from substrate -> air (your current convention)
        "air_to_substrate" means z increases from air -> substrate
    sigma:
        Interface roughness applied BETWEEN the microslabs you create.
        If your profile already has smooth transitions (e.g., from smooth_profile()),
        sigma should usually be 0.0.
    dq_over_q or q_resolution:
        Use dq_over_q for generic smearing.
        Use q_resolution to match your generator: dq = (q_resolution/2.355) * Q
    """

    # --- Parse input ---
    z = np.asarray(sld_arr[0], dtype=float)
    rho = np.asarray(sld_arr[1], dtype=float)
    irho = None
    if len(sld_arr) >= 3 and sld_arr[2] is not None:
        irho = np.asarray(sld_arr[2], dtype=float)

    if z.ndim != 1 or rho.ndim != 1 or z.size != rho.size:
        raise ValueError("z and rho must be 1D arrays of equal length")

    # Ensure strictly increasing and unique z (your caller already does this, but keep robust)
    z, idx = np.unique(z, return_index=True)
    rho = rho[idx]
    if irho is not None:
        irho = irho[idx]

    if z.size < 3:
        raise ValueError("Need at least 3 z points to compute reflectivity from a profile")

    if not np.all(np.diff(z) > 0):
        raise ValueError("z must be strictly increasing")

    # --- Reorder to fronting -> backing (air/water -> substrate) ---
    # refl1d assumes neutrons come from the *fronting* medium.
    if order == "substrate_to_air":
        # input is substrate -> air, reverse to air -> substrate
        z = z[::-1]
        rho = rho[::-1]
        if irho is not None:
            irho = irho[::-1]
    elif order == "air_to_substrate":
        pass
    else:
        raise ValueError("order must be 'substrate_to_air' or 'air_to_substrate'")

    # Shift so fronting starts at z=0
    z = z - z[0]

    # --- Optional resampling for stable microslicing ---
    if dz is not None:
        dz = float(dz)
        if dz <= 0:
            raise ValueError("dz must be positive")
        z_fine = np.arange(z[0], z[-1] + 0.5*dz, dz)
        rho = np.interp(z_fine, z, rho)
        if irho is not None:
            irho = np.interp(z_fine, z, irho)
        z = z_fine

    # --- Infer fronting/backing from plateaus if not provided ---
    k = min(int(plateau_pts), z.size)
    if rho_fronting is None:
        rho_fronting = float(np.mean(rho[:k]))
    if rho_backing is None:
        rho_backing = float(np.mean(rho[-k:]))

    irho_fronting = 0.0 if irho is None else float(np.mean(irho[:k]))
    irho_backing  = 0.0 if irho is None else float(np.mean(irho[-k:]))

    # --- Pad semi-infinite media as thick plateaus (improves low-Q consistency) ---
    # front pad at z<0 and back pad at z>zmax
    if pad_front and pad_front > 0:
        z = np.r_[0.0, z + pad_front]
        rho = np.r_[rho_fronting, rho]
        if irho is not None:
            irho = np.r_[irho_fronting, irho]
        else:
            irho = None  # keep None

    if pad_back and pad_back > 0:
        z = np.r_[z, z[-1] + pad_back]
        rho = np.r_[rho, rho_backing]
        if irho is not None:
            irho = np.r_[irho, irho_backing]

    # Re-shift to start at 0
    z = z - z[0]

    # --- Build Q ---
    if Q is None:
        Q = np.linspace(qmin, qmax, int(n_q))
    else:
        Q = np.asarray(Q, dtype=float)

    # --- Build dQ for convolution ---
    if q_resolution is not None:
        # matches your generator
        dq_over_q_eff = float(q_resolution) / 2.355
    elif dq_over_q is not None:
        dq_over_q_eff = float(dq_over_q)
    else:
        dq_over_q_eff = 0.0

    dQ = np.clip(Q * dq_over_q_eff, 1e-12, None)

    # --- Convert continuous profile to microslabs correctly ---
    w_film = np.diff(z)                      # length N-1
    rho_mid = 0.5 * (rho[:-1] + rho[1:])     # length N-1
    if irho is None:
        irho_mid = np.zeros_like(rho_mid)
    else:
        irho_mid = 0.5 * (irho[:-1] + irho[1:])

    # Now build slab arrays INCLUDING semi-infinite front/back (0 thickness)
    w = np.r_[0.0, w_film, 0.0]              # length (N-1)+2
    rrho = np.r_[rho_fronting, rho_mid, rho_backing]
    iirho = np.r_[irho_fronting, irho_mid, irho_backing]

    # --- Roughness array per interface ---
    n_interfaces = len(rrho) - 1
    if np.isscalar(sigma):
        sigma_arr = np.full(n_interfaces, float(sigma))
    else:
        sigma_arr = np.asarray(sigma, dtype=float)
        if sigma_arr.size != n_interfaces:
            raise ValueError("sigma must be scalar or length len(rrho)-1")

    # --- Reflectivity ---
    # refl1d's low-level routine expects kz (1/Å), not Q (1/Å).
    # For specular reflectometry, Q = 2*kz.
    kz = Q / 2.0
    r = reflamp(kz, w, rrho, iirho, sigma_arr)
    R = np.abs(r) ** 2

    # --- Instrument resolution convolution (only if you want it) ---
    if apply_convolution and dq_over_q_eff > 0:
        R = convolve(Q, R, Q, dQ)

    return Q, R


def compute_nr_from_sld_refnx(
    sld_arr,
    qmin=0.00843,
    qmax=0.09275,
    n_q=400,
    dq_over_q=0.025,
    sigma=3.0,
):
    """
    Compute neutron reflectivity using refnx from a continuous SLD profile.

    Parameters
    ----------
    sld_arr : tuple
        (z, rho) or (z, rho, irho)
        z, rho ordered from substrate -> air
    qmin, qmax : float
        Q range (Å^-1)
    n_q : int
        Number of Q points
    dq_over_q : float
        Fractional resolution ΔQ / Q
    sigma : float
        Interfacial roughness in Å

    Returns
    -------
    Q : ndarray
        Momentum transfer (Å^-1)
    R : ndarray
        Reflectivity
    """

    # -----------------------------
    # Parse input
    # -----------------------------
    z = np.asarray(sld_arr[0], dtype=float)
    rho = np.asarray(sld_arr[1], dtype=float)

    if z.ndim != 1 or rho.ndim != 1 or z.size != rho.size:
        raise ValueError("z and rho must be 1D arrays of equal length")

    if not np.all(np.diff(z) > 0):
        raise ValueError("z must be strictly increasing (substrate → air)")

    # -----------------------------
    # Reverse: substrate → air  →  air → substrate
    # -----------------------------
    z = z[::-1]
    rho = rho[::-1]

    # shift so fronting interface starts at z = 0
    z = z - z[0]

    # -----------------------------
    # Define fronting & backing
    # -----------------------------
    air = SLD(0.0, name="air")
    substrate = SLD(rho[-1], name="substrate")

    # -----------------------------
    # Build slab layers
    # -----------------------------
    layers = []
    for i in range(len(z) - 1):
        thickness = z[i + 1] - z[i]
        mid_sld = 0.5 * (rho[i] + rho[i + 1])

        layer = SLD(mid_sld, name=f"layer{i}")(thickness, sigma)
        layers.append(layer)

    # -----------------------------
    # Build structure: air | film | substrate
    # -----------------------------
    structure = air(0, sigma)
    for layer in layers:
        structure |= layer
    structure |= substrate(0, sigma)

    # -----------------------------
    # Reflectivity model
    # -----------------------------
    model = ReflectModel(structure, bkg=0.0, scale=1.0)
    model.dq = dq_over_q

    # -----------------------------
    # Compute NR
    # -----------------------------
    Q = np.linspace(qmin, qmax, n_q)
    R = model(Q)

    return Q, R


# for experimental data
def reverse_y_order(sld_array):
    """
    Flip SLD y from left to right, substrate to air direction
    """
    A_flipped = sld_array.copy()
    A_flipped[1] = A_flipped[1, ::-1]
    return A_flipped


def find_substrate_critical_idx(arr, target=2.075):
    """
    the idx of critical point where substrate transit to material film
    """
    y = arr[1]
    for i in range(len(y) - 1):
        if y[i] <= target and y[i + 1] > target:
            return i


def align_points(A, B):
    """
    A is expt sld, B is pred sld

    Align B to A using shift.
    Both A and B are shape (2, N).
    """
    # Step 1: find index in A where y ≈ 2.07 (left edge)
    target_y = 2.07

    # Search from left to right, the closest idx x which y close to target y
    idx_A = find_substrate_critical_idx(A)

    idx_B = find_substrate_critical_idx(B)

    # Step 2: compute the translation vector
    shift = A[:, idx_A] - B[:, idx_B]

    # Step 3: apply the shift
    B_aligned = B + shift[:, np.newaxis]

    return B_aligned



from pathlib import Path
import typer

from ..models.config import NRSLDCurvesGeneratorParams,NRSLDModelTrainerParams,NRSLDModelInferenceParams
from . import reflectivity_pipeline as workflow
from .reflectivity_pipeline import load_normalization_stat

from ..config.errors import ConfigMissingKeyError
from ..input import NRSLDDataProcessor

def run_sld_prediction(root:str | Path,config:dict):
    config["root"] = root
    model_path = root / Path(config["nr_predict_sld"]["models"]["model"])

    # Train Model and save if no model found
    if not model_path.exists():
        typer.echo("No trained SLD model found. Training a new model...")

        #Generate curves for model training and save generated data to path
        try:
            # Init data generator params from config
            generator_params = NRSLDCurvesGeneratorParams(_config=config)

        except ConfigMissingKeyError as e:
            typer.echo(e)
            raise typer.Exit()

        # Prepare synthetic data
        typer.echo("Generating Curves for training...")
        workflow.generate_nr_sld_curves(generator_params)

        # Init training params from config
        trainer_params = NRSLDModelTrainerParams(_config=config)

        #Preprocessing data
        dproc = NRSLDDataProcessor(trainer_params.nr_file,trainer_params.sld_file).load_data()

        X,y = workflow.preprocess(dproc,trainer_params.normalization_stats)

        typer.echo("Training model using curves...")

        #Train and save
        workflow.train_nr_predict_sld_model(X,y,trainer_params, auto_save=True)

    # Load model
    model = workflow.load_nr_sld_model(model_path)
    typer.echo("Loaded NR SLD model...")


    inference_params = NRSLDModelInferenceParams(_config=config)

    norm_stats = load_normalization_stat(inference_params.normalization_stats)

    #Inference
    predicted_sld = workflow.predict_sld_from_nr(
        model,
        inference_params.experimental_nr,
        norm_stats)

    return predicted_sld



import torch
import pandas as pd
from . import autoencoder as ae
from ..input.data_processor import DataProcessor
from . import mlp as mlp_module
from ..config.runtime import DEVICE as device

class ChiPredModelTrainer:
    def __init__(self, autoencoder, mlp, batch_size, ae_epochs, mlp_epochs, loss_fn, latent_dim,
                 num_params):
        self.autoencoder = autoencoder
        self.mlp = mlp
        self.batch_size = batch_size
        self.ae_epochs = ae_epochs
        self.mlp_epochs = mlp_epochs
        self.loss_fn = loss_fn
        self.latent_dim = latent_dim
        self.num_params = num_params
        self.trained = False  # Track if models are trained

    def train_autoencoder(self, tr_load, val_load):
        """Train the autoencoder model and store losses."""
        print("\nTraining Autoencoder...")
        self.ae_train_loss, self.ae_val_loss = ae.train(self.autoencoder, tr_load, val_load,
                                                                      self.ae_epochs, self.loss_fn)
        self.trained = True

    def extract_latent_vectors(self, tr_data):
        """Extract latent vectors from the trained Autoencoder and store them in a Pandas DataFrame."""
        assert self.trained, "Autoencoder must be trained before extracting latent vectors."

        print("\nExtracting Latent Vectors...")
        encoded_samples = []
        self.autoencoder.eval()

        with torch.no_grad():
            for sample in tr_data:
                curve = sample[0].view(sample[0].size(0), -1).flatten().to(device)
                labels = sample[1]
                latent_vars = self.autoencoder.encoder(curve).cpu().numpy()

                latent_sample = {f"l{i + 1}": var for i, var in enumerate(latent_vars)}
                for i in range(self.num_params):
                    label_index = f'chi{i + 1}'
                    latent_sample[label_index] = labels[i].item()

                encoded_samples.append(latent_sample)

        return pd.DataFrame(encoded_samples)

    def get_full_output(self, ae_model, dataloader):
        """Run the Autoencoder on a dataset and return tensors for curves, reconstructions, latent space, and parameters."""
        list_curves, list_recon, list_latent, list_labels = [], [], [], []

        ae_model.to(device)
        ae_model.eval()
        with torch.no_grad():
            for curve, labels in dataloader:
                curve, labels = curve.to(device), labels.to(device)  # Move data to the correct device

                latent_vars = ae_model.encoder(curve.view(curve.size(0), -1))
                recon_curve = ae_model.decoder(latent_vars)

                list_curves.append(curve)
                list_labels.append(labels)
                list_latent.append(latent_vars)
                list_recon.append(recon_curve)

        return [
            torch.cat(list_curves).to(device),
            torch.cat(list_recon).to(device),
            torch.cat(list_latent).to(device),
            torch.cat(list_labels).to(device),
        ]

    def prepare_mlp_data(self, tr_load, val_load, tst_load):
        """Prepare MLP input datasets using latent space vectors and chi parameters."""
        print("\nPreparing MLP Training Data...")
        full_data_train = self.get_full_output(self.autoencoder, tr_load)
        full_data_valid = self.get_full_output(self.autoencoder, val_load)
        full_data_test = self.get_full_output(self.autoencoder, tst_load)

        mlp_input_data = [
            full_data_train[2], full_data_train[3],
            full_data_valid[2], full_data_valid[3],
            full_data_test[2], full_data_test[3]
        ]

        return DataProcessor.get_dataloaders(*mlp_input_data, self.batch_size)

    def train_mlp(self, mlp_tr_load, mlp_val_load):
        """Train the MLP model and store losses."""
        print("\nTraining MLP...")
        self.train_loss, self.val_loss = mlp_module.train(self.mlp, mlp_tr_load, mlp_val_load, self.mlp_epochs,
                                                        self.loss_fn)

    def evaluate_model(self, mlp_data, model):
        """Run the trained MLP on a dataset and store predictions in a Pandas DataFrame."""
        print("\nEvaluating MLP Model...")
        samples = []
        model.eval()

        with torch.no_grad():
            for sample in mlp_data:
                img = sample[0].view(sample[0].size(0), -1).flatten().to(device)
                label_val = sample[1]

                out_label = model(img).cpu().numpy()
                sample_dict = {f"pred_chi{i + 1}": enc for i, enc in enumerate(out_label)}

                for i in range(self.num_params):
                    label_index = f'chi{i + 1}'
                    sample_dict[label_index] = float(label_val[i])

                samples.append(sample_dict)

        return pd.DataFrame(samples)

    def calculate_error(self, df_samples):
        """Calculate percentage error in chi predictions."""
        print("\nCalculating Prediction Errors...")
        for i in range(self.num_params):
            df_samples[f'chi{i + 1}_err'] = (abs(df_samples[f'pred_chi{i + 1}'] - df_samples[f'chi{i + 1}']) /
                                             df_samples[f'chi{i + 1}']) * 100

        df_errors = pd.DataFrame(
            {f'chi{i + 1}_err': [df_samples[f'chi{i + 1}_err'].mean()] for i in range(self.num_params)})
        return df_errors



from pathlib import Path
from typing import Optional, Tuple

import numpy as np
import torch
import typer

from ..config.runtime import DEVICE
from ..input.data_processor import DataProcessor, NRSLDDataProcessor
from ..input.reflectivity_data_generator import ReflectivityDataGenerator
from ..models.cnn import CNN
from ..models.config import (
    NRSLDCurvesGeneratorParams,
    NRSLDModelTrainerParams,
)
from ..models.nr_sld_model_trainer import NRSLDModelTrainer


def _compute_norm_stats(curves: np.ndarray) -> dict:
    """Return min/max stats for x and y dimensions."""

    x_points = curves[:, 0, :]
    y_points = curves[:, 1, :]
    return {
        "x": {"min": float(np.min(x_points)), "max": float(np.max(x_points))},
        "y": {"min": float(np.min(y_points)), "max": float(np.max(y_points))},
    }


class ReflectivityPipeline:
    """High level orchestration for the NR->SLD workflow."""

    def __init__(
        self,
        generator_params: Optional[NRSLDCurvesGeneratorParams] = None,
        trainer_params: Optional[NRSLDModelTrainerParams] = None,
    ):
        self.generator_params = generator_params
        self.trainer_params = trainer_params

    def generate_synthetic_curves(
        self,
        layer_desc=None,
        layer_bound=None,
    ) -> Tuple[np.ndarray, np.ndarray]:
        if self.generator_params is None:
            raise ValueError("generator_params must be provided to generate curves.")

        return generate_nr_sld_curves(
            self.generator_params,
            layer_desc=layer_desc,
            layer_bound=layer_bound,
        )

    def prepare_training_data(self) -> Tuple[np.ndarray, np.ndarray]:
        if self.trainer_params is None:
            raise ValueError("trainer_params must be provided to preprocess data.")

        dproc = NRSLDDataProcessor(
            self.trainer_params.nr_file,
            self.trainer_params.sld_file,
        ).load_data()

        return preprocess(dproc, self.trainer_params.normalization_stats)

    def train_model(
        self,
        dataset: Optional[Tuple[np.ndarray, np.ndarray]] = None,
        auto_save: bool = True,
    ) -> torch.nn.Module:
        if self.trainer_params is None:
            raise ValueError("trainer_params must be provided to train a model.")

        if dataset is None:
            dataset = self.prepare_training_data()

        reshaped_nr, normalized_sld = dataset
        return train_nr_predict_sld_model(
            reshaped_nr,
            normalized_sld,
            self.trainer_params,
            auto_save=auto_save,
        )

    def load_model(self, model_path: Optional[str | Path] = None) -> torch.nn.Module:
        if model_path is None:
            if self.trainer_params is None:
                raise ValueError("trainer_params required to infer model path.")
            model_path = self.trainer_params.model_path
            layers = self.trainer_params.layers
            dropout = self.trainer_params.dropout
        else:
            layers = self.trainer_params.layers if self.trainer_params else None
            dropout = self.trainer_params.dropout if self.trainer_params else None

        return load_nr_sld_model(model_path, layers=layers, dropout_prob=dropout)

    def predict_sld(
        self,
        nr_data: np.ndarray | str | Path,
        norm_stats: Optional[dict] = None,
        model: Optional[torch.nn.Module] = None,
    ) -> np.ndarray:
        if model is None:
            model = self.load_model()

        if norm_stats is None:
            if self.trainer_params is None:
                raise ValueError("Normalization statistics are required for prediction.")
            norm_path = self.trainer_params.normalization_stats
            norm_stats = load_normalization_stat(norm_path)

        return predict_sld_from_nr(model, nr_data, norm_stats)


def generate_nr_sld_curves(
    params: NRSLDCurvesGeneratorParams,
    layer_desc=None,
    layer_bound=None,
) -> Tuple[np.ndarray, np.ndarray]:
    """Generate NR/SLD pairs and persist them via the generator params."""

    data_generator = ReflectivityDataGenerator(
        num_layers=params.num_film_layers,
        layer_desc=layer_desc,
        layer_bound=layer_bound,
    )

    nr_curves, sld_curves = data_generator.generate(params.num_curves)

    nr_path = Path(params.mod_nr_file)
    sld_path = Path(params.mod_sld_file)
    nr_path.parent.mkdir(parents=True, exist_ok=True)
    sld_path.parent.mkdir(parents=True, exist_ok=True)

    np.save(nr_path, nr_curves)
    np.save(sld_path, sld_curves)

    typer.echo(
        "Synthetic NR/SLD curves generated and saved at:"
        f"\n  NR curves: {nr_path}\n  SLD curves: {sld_path}"
    )

    return nr_curves, sld_curves


def preprocess(
    dproc: NRSLDDataProcessor,
    norm_stats_to_be_save_path: str | Path,
) -> Tuple[np.ndarray, np.ndarray]:
    """
    Preprocess function normalization and remove Q from NR.
    :param dproc: DataProcessor, get nr,sld from load data
    :return: nr, sld
    """
    nr_curves = getattr(dproc, "_nr_arr", None)
    sld_curves = getattr(dproc, "_sld_arr", None)

    if nr_curves is None or sld_curves is None:
        raise ValueError("DataProcessor must load NR and SLD arrays before preprocessing.")

    nr_curves = np.array(nr_curves, copy=True)
    nr_curves[:, 1, :] = np.log10(np.clip(nr_curves[:, 1, :], 1e-8, None))
    nr_stats = _compute_norm_stats(nr_curves)
    normalized_nr = DataProcessor.normalize_xy_curves(
        getattr(dproc, "_nr_arr"),
        apply_log=True,
        min_max_stats=nr_stats,
    )

    sld_stats = _compute_norm_stats(sld_curves)
    normalized_sld = DataProcessor.normalize_xy_curves(
        sld_curves,
        apply_log=False,
        min_max_stats=sld_stats,
    )

    norm_stats = {"nr": nr_stats, "sld": sld_stats}
    norm_stats_path = Path(norm_stats_to_be_save_path)
    norm_stats_path.parent.mkdir(parents=True, exist_ok=True)
    np.save(norm_stats_path, norm_stats, allow_pickle=True)

    reshaped = dproc.reshape_nr_to_single_channel(normalized_nr)

    return reshaped, normalized_sld

def load_normalization_stat(norm_stat_path) -> dict:
    """
    Load min max for normalization used in training from file.
    :param norm_stat_path:
    :return: min max dict of nr and sld
    """
    return np.load(norm_stat_path, allow_pickle=True).item()

def train_nr_predict_sld_model(
    reshaped_nr_curves,
    normalized_sld_curves,
    params: NRSLDModelTrainerParams,
    auto_save=True,
) -> torch.nn.Module:
    """
    Model training
    :param reshaped_nr_curves:
    :param normalized_sld_curves:
    :param params:
    :param auto_save: boolean
    :return: torch.nn.Module
    """

    trainer = NRSLDModelTrainer(
        X=reshaped_nr_curves,
        y=normalized_sld_curves,
        layers=params.layers,
        batch_size=params.batch_size,
        epochs=params.epochs,
        dropout=params.dropout,
    )

    # training
    model = trainer.train_pipeline()

    # save model
    if auto_save:
        to_be_saved_model_path = params.model_path
        torch.save(model.state_dict(), to_be_saved_model_path)
        typer.echo(
            "NR predict SLD trained CNN model saved at: "
            f"{to_be_saved_model_path}"
        )

    return model

def load_nr_sld_model(
    model_path: str | Path,
    *,
    layers: Optional[int] = None,
    dropout_prob: Optional[float] = None,
) -> torch.nn.Module:
    """Load a trained CNN checkpoint for NR->SLD prediction."""

    model_layers = layers or 12
    model_dropout = dropout_prob if dropout_prob is not None else 0.5

    model = CNN(layers=model_layers, dropout_prob=model_dropout)
    state_dict = torch.load(Path(model_path), map_location=DEVICE)
    model.load_state_dict(state_dict)
    typer.echo(f"Loaded NR predict SLD model from: {model_path}")

    return model


def predict_sld_from_nr(model, nr_data: np.ndarray | Path | str, norm_stats) -> np.ndarray:
    """
    Predicts SLD profiles from given NR curves using a trained model.

    :param model:torch.nn.Module: trained CNN model
    :param nr_data: np.ndarray | str: NR data
    :param norm_stats: dict: normalization status
    :return: np.ndarray: SLD profiles
    """
    norm_stats_nr = norm_stats["nr"]
    norm_stats_sld = norm_stats["sld"]
    try:
        if isinstance(nr_data, (str,Path)):
            processor = NRSLDDataProcessor(nr_file_path=nr_data).load_data()
            norm_nr = processor.normalize_nr(norm_stats_nr)
        elif isinstance(nr_data, np.ndarray):
            norm_nr = DataProcessor.normalize_xy_curves(
                nr_data,
                apply_log=True,
                min_max_stats=norm_stats_nr,
            )
            processor = NRSLDDataProcessor()
        else:
            raise TypeError("nr_data must be a file path (str) or a NumPy array.")
    except FileNotFoundError as e:
        typer.echo(f"Error loading NR data: {e}")
        raise typer.Exit()

    # Remove wave vector (x channel) of NR
    reshaped = processor.reshape_nr_to_single_channel(norm_nr)

    y = _predict(model, reshaped)

    predicted_sld_curves = processor.denormalize(
        y,
        curve_type='sld',
        min_max_stats=norm_stats_sld)

    typer.echo(f"Predicted SLD shape: {predicted_sld_curves.shape}")

    return predicted_sld_curves

def _predict(model, X_batch:np.ndarray)->np.ndarray:
    model.eval().to(DEVICE)
    X_batch = torch.tensor(X_batch, dtype=torch.float32).to(DEVICE)

    with torch.no_grad():
        y = model(X_batch)

    return y.cpu().numpy()



#This block is the class that is used to generate the SLD and NR curves
from typing import List, Tuple

import typer
from tqdm.auto import tqdm
from scipy.stats import qmc
# from scipy.interpolate import CubicSpline
from refl1d.names import *
import numpy as np
import os

class ReflectivityDataGenerator:
    """
        Generate pairs of NR, SLD profile curves using Reflectivity Model
    """
    def __init__(self,num_layers,layer_desc = None, layer_bound = None):
        self.num_layers = num_layers
        match self.num_layers:
            # Use physical model from original code
            case 5 if layer_desc is None and layer_bound is None:
                typer.echo(f"Using Legacy Reflectivity Model layer{num_layers}...\n")
                self.model = ReflectivityFiveLayerModel()
            case _:
                typer.echo(f"Using Model layer{num_layers}...\n")
                self.model = ReflectivityModel(num_layers,layer_desc=layer_desc,layer_bound=layer_bound)

    def generate(self,num_curves:int)-> Tuple[np.ndarray, np.ndarray]:
        """
        For each curve, sample a random set of parameters
            within the defined bounds for each layer,
            update the fixed model with these parameters, and
            pass the resulting model into calculate_reflectivity
            to compute the reflectivity curve.

        :param num_curves: int
        :return Tuple(np.ndarray,np.ndarray):
        """
        self.model.generate(num_curves)
        #add noise and log transform NR
        # pars, train_data = self.model.get_preprocessed_data()

        #standardize all sld profiles to start at same place
        for sld_arr in self.model._smooth_array:
            sld_arr[0] = np.array(sld_arr[0]) - np.min(sld_arr[0])

        nr_curves = []
        sld_curves = []
        for i in range(len(self.model._smooth_array)):
            nr_curves.append(np.array([self.model.q,self.model._refl_array[i]]))
            sld_curves.append(np.array(self.model._smooth_array[i]))

        return np.stack(nr_curves), np.stack(sld_curves)


def calculate_reflectivity(q, model_description, q_resolution=0.0294855):
    """
        Reflectivity calculation using refl1d
    """
    zeros = np.zeros(len(q))
    dq = q_resolution * q / 2.355

    # The QProbe object represents the beam
    probe = QProbe(q, dq, data=(zeros, zeros))

    layers = model_description['layers']
    sample = Slab(material=SLD(name=layers[0]['name'],
                               rho=layers[0]['sld']), interface=layers[0]['roughness'])
    # Add each layer
    for l in layers[1:]:
        sample = sample | Slab(material=SLD(name=l['name'],
                               rho=l['sld'], irho=l['isld']),
                               thickness=l['thickness'], interface=l['roughness'])

    probe.background = Parameter(value=model_description['background'], name='background')
    expt = Experiment(probe=probe, sample=sample)

    q, r = expt.reflectivity()
    z, sld, _ = expt.smooth_profile()
    #this makes all SLD curves have 900 datapoints per curve
    zNew = np.linspace(z[0], z[-1], num=900)
    sldNew = np.interp(zNew, z, sld)  # linear, no overshoot
    return model_description['scale'] * r, zNew, sldNew


class ReflectivityFiveLayerModel(object):
    """
    Reflectivity model with five layers, with substrate, and SiOx layer as base,
    polymer film layer in the middle and top air layer.
    """
        # Neutrons come in from the last item in the list
    def __init__(self,q=None, name='polymer'):

        self.parameters = [
                  dict(i=0, par='roughness', bounds=[1.177, 1.5215]),
                  # The following is the Si oxide layer
                  dict(i=1, par='sld', bounds=[3.47, 3.47]),
                  dict(i=1, par='thickness', bounds=[9.7216, 14.624]),
                  dict(i=1, par='roughness', bounds=[1.108, 2.998]),

                  # The next 5 layers are the polymer
                    dict(i=2, par='sld', bounds=[3.7235, 4.197]),
                  dict(i=2, par='thickness', bounds=[8.717,98.867]),
                  dict(i=2, par='roughness', bounds=[2.2571,38.969]),
                    dict(i=3, par='sld', bounds=[1.6417, 3.1033]),
                  dict(i=3, par='thickness', bounds=[117.4,239.91]),
                  dict(i=3, par='roughness', bounds=[19.32,95.202]),
                    dict(i=4, par='sld', bounds=[3.0246, 4.6755]),
                  dict(i=4, par='thickness', bounds=[64.482,94.768]),
                  dict(i=4, par='roughness', bounds=[15.713,28.007]),
                  dict(i=5, par='sld', bounds=[1.501, 4.9837]),
                  dict(i=5, par='thickness', bounds=[51.655,83.334]),
                  dict(i=5, par='roughness', bounds=[9.7741,25.373]),
                  dict(i=6, par='sld', bounds=[0.85516,4.4906]),
                  dict(i=6, par='thickness', bounds=[58.479, 86.738]),
                  dict(i=6, par='roughness', bounds=[43.155, 110.11]),
                 ]
        self.model_description = dict(layers=[
                                dict(sld=2.07, isld=0, thickness=0, roughness=1.8272, name='substrate'),
                                dict(sld=3.47, isld=0, thickness=10.085, roughness=1.108, name='siox'),
                                dict(sld=3.734, isld=0, thickness=8.717, roughness=2.2571, name='bottom'),
                                dict(sld=3.1033, isld=0, thickness=239.91, roughness=37.707, name='bottom'),
                                dict(sld=3.0246, isld=0, thickness=91.232, roughness=20.147, name='bulk'),
                                dict(sld=3.0246, isld=0, thickness=51.655, roughness=20.147, name='bulk'),
                                dict(sld=0.85516, isld=0, thickness=62.189, roughness=43.155, name='bottom'),
                                dict(sld=0, isld=0, thickness=0, roughness=0, name='air')
                         ],
                         scale=1,
                         background=0,
                        )
        # The following are unmodified physical parameters and corresponding reflectivity data
        self._pars_array = []
        self._refl_array = []
        # This creates the SLD curve list
        self._smooth_array = []
        # The following are the parameters, mapped between -1 and 1.
        self._train_pars = []
        self._train_data = None
        self._config_name = name

        if q is None:
            self.q = np.logspace(np.log10(0.008101436040354381), np.log10(0.1975709062238298), num=308)
        else:
            self.q = q

    @classmethod
    def from_dict(cls, pars, q_array=None):
        """
            Create ReflectivityModels object from a dict that
            defines the reflectivity model parameters and how
            the training set should be generated.
        """
        m = cls(q_array, name=pars['name'])
        m.model_description =  pars['model']
        m.parameters = pars['parameters']
        return m

    def generate(self, n=100):
        """
            For each curve, sample a random set of parameters
            within the defined bounds for each layer,
            update the fixed model with these parameters, and
            pass the resulting model into calculate_reflectivity
            to compute the reflectivity curve.
        """
        npars = len(self.parameters)
        self._train_pars = np.random.uniform(low=-1, high=1, size=[n, npars])
        # Compute model parameters and reflectivity using these values
        self.compute_reflectivity()

    def to_model_parameters(self, pars):
        """
            Transform an array of parameters to a list of calculable models
        """
        pars_array = np.zeros(pars.shape)

        for i, par in enumerate(self.parameters):
            a = (par['bounds'][1]-par['bounds'][0])/2.
            b = (par['bounds'][1]+par['bounds'][0])/2.
            pars_array.T[i] = pars.T[i] * a + b

        return pars_array

    def compute_reflectivity(self):
        """
            Transform an array of parameters to a list of calculable models
            and compute reflectivity
        """
        print("Computing reflectivity")
        self._pars_array = self.to_model_parameters(self._train_pars)
        # Compute reflectivity
        for p in tqdm(self._pars_array,desc="Generating reflectivity curves",colour="green"):
            _desc = self.get_model_description(p)
            r, z, sld = calculate_reflectivity(self.q, _desc)
            self._refl_array.append(r)
            self._smooth_array.append([z,sld])

    def get_model_description(self, pars):
        """
            Convert the parameter list to a model description that we can use
            to compute R(q).
        """
        for i, par in enumerate(self.parameters):
            self.model_description['layers'][par['i']][par['par']] = pars[i]
        return self.model_description

    def get_preprocessed_data(self, errors=None):
        """
            Pre-process data
            If errors is provided, a random error will be added, taking the errors array
            as a relative uncertainty.
        """
        if errors is None:

            self._train_data = np.log10(self._refl_array*self.q**2/self.q[0]**2)
        else:
            _data = self._refl_array * (1.0 + np.random.normal(size=len(errors)) * errors)
            # Catch the few cases where we generated a negative intensity and take
            # the absolute value
            _data[_data<0] = np.fabs(_data[_data<0])
            self._train_data = np.log10(_data*self.q**2/self.q[0]**2)

        return self._train_pars, self._train_data

    def save(self, output_dir=''):
        """
            Save all data relevant to a training set
            @param output_dir: directory used to store training sets
        """
        # Save q values
        np.save(os.path.join(output_dir, "%s_q_values" % self._config_name), self.q)

        # Save training set
        if self._train_data is not None:
            np.save(os.path.join(output_dir, "%s_data" % self._config_name), self._train_data)
            np.save(os.path.join(output_dir, "%s_pars" % self._config_name), self._train_pars)

    def load(self, data_dir=''):
        self.q = np.load(os.path.join(data_dir, "%s_q_values.npy" % self._config_name))
        self._train_data = np.load(os.path.join(data_dir, "%s_data.npy" % self._config_name))
        self._train_pars = np.load(os.path.join(data_dir, "%s_pars.npy" % self._config_name))
        return self.q, self._train_data, self._train_pars


class ReflectivityModel:
    """
        A revised version of the reflectivity model that has fixed 6 layers. This model support dynamic input layers.
    """
    def __init__(self,
                 num_layers: int,
                 layer_desc: List[dict] = None,
                 layer_bound: List[dict] = None,
                 q=None):

        self.q = np.asarray(q) if q is not None else np.logspace(np.log10(0.0081), np.log10(0.1975), num=308)
        self.num_layers = num_layers
        self.model_description = {
            "layers": layer_desc if layer_desc is not None else self._auto_generate_layer_description(),
            "scale": 1,
            "background": 0,
        }
        self.parameters = layer_bound if layer_bound is not None else self._auto_generate_layer_bounds()

        # The following are unmodified physical parameters and corresponding reflectivity data
        self._pars_array = []
        self._refl_array = []
        # This creates the SLD curve list
        self._smooth_array = []
        # The following are the parameters, mapped between -1 and 1.
        self._train_pars = []
        self._train_data = None

    def _auto_generate_layer_bounds(self):
        typer.echo("Generating random layer bounds...")
        bounds=[
            dict(i=0, par="roughness", bounds=[0, 5]),
            #siox layer
            dict(i=1,par="sld",bounds=[3,3.9]),
            dict(i=1,par="thickness",bounds=[10,25]),
            dict(i=1,par="roughness",bounds=[1,10])
        ]

        for i in range(self.num_layers):
            idx = i + 2  # layer indices after substrate and siox
            bounds.extend([
                dict(i=idx, par='sld', bounds=[0, 6.4]),
                dict(i=idx, par='thickness', bounds=[10, 300]),
                dict(i=idx, par='roughness', bounds=[0, 150])
            ])
        return bounds

    def _auto_generate_layer_description(self):
        typer.echo("Generating layer description based on layer bounds...")

        random_layers = [
                     dict(sld=2.07, isld=0, thickness=0, roughness=3, name='substrate'),
                     dict(sld=3.3, isld=0, thickness=16.2139, roughness=7, name='siox'),
                 ] + [
                     dict(
                         sld=np.random.uniform(0, 6.4),
                         isld=0,
                         thickness=np.random.randint(10, 300),
                         roughness=np.random.randint(0, 150),
                         name=f"layer{i}"
                     ) for i in range(self.num_layers)
                 ] + [
                     dict(sld=0.0, isld=0, thickness=0, roughness=0, name='air')
                 ]

        return random_layers

    def generate(self, n=100):
        npars = len(self.parameters)

        sampler = qmc.LatinHypercube(d=npars)
        sample = sampler.random(n=n) # shape: (n, npars)
        # self._train_pars = np.random.uniform(low=-1, high=1, size=[n, npars])
        self._train_pars = sample*2 -1 # map linearly to [-1, 1]
        self.compute_reflectivity()

    def to_model_parameters(self, pars):
        pars_array = np.zeros(pars.shape)
        for i, par in enumerate(self.parameters):
            a = (par['bounds'][1] - par['bounds'][0]) / 2.
            b = (par['bounds'][1] + par['bounds'][0]) / 2.
            pars_array.T[i] = pars.T[i] * a + b
        return pars_array

    def compute_reflectivity(self):
        self._pars_array = self.to_model_parameters(self._train_pars)
        for p in tqdm(self._pars_array,desc="Computing reflectivity for NR, SLD profile",colour='green'):
            _desc = self.get_model_description(p)
            r, z, sld = calculate_reflectivity(self.q, _desc)
            self._refl_array.append(r)
            self._smooth_array.append([z, sld])

    def get_model_description(self, pars):
        for i, par in enumerate(self.parameters):
            self.model_description['layers'][par['i']][par['par']] = pars[i]
        return self.model_description




import torch
from torch import nn

class CNN(nn.Module):
    def __init__(self, layers=12,dropout_prob = 0.5):
        super(CNN, self).__init__()
        self.layers = nn.ModuleList()
        addition = 255 / layers
        curr = 1
        for hdim in range(layers - 1):
            self.layers.append(nn.Conv1d(int(curr + 0.5), int(curr + addition + 0.5), 51, padding=25))
            self.layers.append(nn.BatchNorm1d(int(curr + addition + 0.5)))
            self.layers.append(nn.ReLU(True))
            self.layers.append(nn.Dropout(dropout_prob))
            curr += addition

        self.layers.append(nn.Conv1d(int(curr + 0.5), 256, 51, padding=25))
        self.layers.append(nn.BatchNorm1d(256))
        self.layers.append(nn.ReLU(True))
        self.layers.append(nn.Dropout(dropout_prob))

        self.linear1 = nn.Linear(256 * 308, 900 * 2)

    def forward(self, x):
        for layer in self.layers:
            x = layer(x)
        x = x.reshape(x.shape[0], -1)
        x = self.linear1(x)
        x = x.reshape(-1, 2, 900)
        # (x, 2, 2000)
        return torch.sigmoid(x)


from ..input import DataProcessor

from ..config.runtime import DEVICE
import torch
from .cnn import CNN

from .model_trainer import ModelTrainer

# Training parameters
LEARNING_RATE = 2.15481e-05
WEIGHT_DECAY = 2.6324e-05
SPLIT_RATIO = 0.8

class NRSLDModelTrainer(ModelTrainer):
    def __init__(self,X,y,layers, batch_size, epochs,dropout=0.5):
        super().__init__(batch_size = batch_size, epochs = epochs)
        self.model = CNN(layers=layers,dropout_prob=dropout).to(DEVICE) #model
        self.X = X
        self.y = y
        self.layers = layers
        self.batch_size = batch_size
        self.epochs = epochs

    def train_pipeline(self):
        self.model.train()

        list_arrays = DataProcessor.split_arrays(self.X, self.y, size_split=SPLIT_RATIO)
        tensor_arrays = DataProcessor.convert_tensors(list_arrays)

        # train valid split
        train_dataset, valid_dataset, test_dataset, train_loader, valid_loader, test_loader = DataProcessor.get_dataloaders(
            *tensor_arrays, batch_size= self.batch_size)

        optimizer = torch.optim.Adam(self.model.parameters(), lr=LEARNING_RATE, weight_decay=WEIGHT_DECAY)
        loss_fn = torch.nn.MSELoss()

        # Training loop
        for epoch in range(self.epochs):
            train_loss = self.train_model(self.model, train_loader, optimizer, loss_fn)
            val_loss = self.validate_model(self.model, valid_loader, loss_fn)
            print(f"Epoch {epoch + 1}/{self.epochs} - Train Loss: {train_loss:.6f}, Validation Loss: {val_loss:.6f}")

        return self.model


